#ifndef UDGSETTINGSPARSER_H
#define UDGSETTINGSPARSER_H

#include "singleton.h"

#include <QMap>

namespace udg {

/**
    Classe encarregada de parsejar strings de settings que poden
    contenir un seguit de paraules clau. Noms tindrem una sola instncia
    ja que es tracta d'una classe Singleton.

    Les paraules clau disponibles es carreguen inicialment al crear l'objecte
    conjuntament amb els seus corresponents valors.

    El catleg de paraules clau s el segent:

    %HOSTNAME%: Nom de la mquina local (aka localHostName)
    %IP%: Adrea ip de la mquina local en format aaa.bbb.ccc.dddd
    %IP.1%: Primer prefix de l'adrea ip de la mquina local
    %IP.2%: Segon prefix de l'adrea ip de la mquina local
    %IP.3%: Tercer prefix de l'adrea ip de la mquina local
    %IP.4%: Quart prefix de l'adrea ip de la mquina local
    %USERNAME%: Nom d'usuari actual
    %HOMEPATH%: Ruta al directori "home" d'usuari

    El seu tractament ser "case sensitive".

    A ms a ms, a aquestes paraules clau se'ls pot afegir un sufix que els hi apliqui una mscara.

    == Mscara de truncatge ==

    Aquesta mscara el que fa s truncar a 'n' carcters la clau %KEY%. Podem especificar adicionalment
    un carcter de padding amb el que s'omplin el nombre de carcters buits si la longitud de %KEY% s
    inferior al nombre de carcters truncats.

    La sintaxi s la segent

    %KEY%[n:c]

    on:
      * %KEY%: paraula clau del catleg (HOSTANAME,IP, etc)
      * n: nombre de carcters a truncar de %KEY%. Ha de ser un nombre natural > 0
      * c: carcter de padding amb el que omplirem els espais buits.
           El carcter de padding pot ser qualsevol carcter, excepte l'espai en blanc

    Exemples:

    Per obtenir els 5 ltims dgits de l'adrea IP i omplir els espais buits amb el carcter 'x', escriurem
    %IP.3%[2:x]%IP.4%[3:x]
    Si tingussim la adrea IP 10.80.9.2 el resultat seria "x9xx2"

    Per obtenir els 4 ltims dgits, sense padding escriurem
    %IP.3%[1:]%IP.4%[3:]
    i en aquest cas, per la mateixa adrea IP que l'anterior, el resultat seria "92"
  */
class SettingsParser : public Singleton<SettingsParser> {
public:
    /// Donada una string la parseja i ens retorna el resultat
    QString parse(const QString &stringToParse);

protected:
    /// Cal declarar-ho friend perqu sin haurem de fer pblics
    /// el constructor i destructor i trencarem aix la filosofia d'un Singleton
    friend class Singleton<SettingsParser>;
    SettingsParser();
    ~SettingsParser();

private:
    /// Inicialitza la taula amb les paraules clau i els seus corresponents valors
    void initializeParseableStringsTable();

    /// Ens diu si la cadena de texte passada s una adrea IPv4 vlida.
    /// Noms accepta com a resultats vlids cadenes de texte que tinguin nicament l'adrea IP
    /// Si la cadena cont una adrea IP a ms d'altres carcters, aquesta ser donada com a invlida.
    /// Es considera una adrea IP vlida una cadena amb la forma xxx.xxx.xxx.xxx on xxx s un nmero
    /// dins del rang 0..255. La cadena 192.168.2.1 seria considerada com a vlida,
    /// per la cadena 192.168.002.001 no ho seria pas.
    /// TODO Aquest mtode hauria de formar part d'una classe amb utilitats de sistema o xarxa
    /// accessible des de qualsevol altre lloc de l'aplicaci per fer aquest tipu de comprovacions.
    bool isIPv4Address(const QString &ipAddress);

    /// Ens retorna una llista amb les adreces IPv4 de la mquina local.
    /// Normalment el primer element de la llista hauria de ser la IP corresponent
    /// a la connexi de xarxa principal.
    /// TODO Aquest mtode hauria de formar part d'una classe amb utilitats de sistema o xarxa
    /// accessible des de qualsevol altre lloc de l'aplicaci per fer obtenir aquest tipu d'informaci.
    QStringList getLocalHostIPv4Addresses();

private:
    /// Mapa que cont la relaci de paraules clau amb el seu valor
    QMap<QString, QString> m_parseableStringsTable;

};

} // End namespace udg

#endif
